@@ -6,7 +6,6 @@ from cftool.misc import *
 
 from .types import *
 from ..types import *
-from .wrapper import TabularData
 
 
 class SplitResult(NamedTuple):
@@ -42,6 +41,7 @@ class DataSplitter(SavingMixin):
     >>> from cfdata.types import np_int_type
     >>> from cfdata.tabular.types import TaskTypes
     >>> from cfdata.tabular.wrapper import TabularDataset
+    >>> from cfdata.tabular.utils import DataSplitter
     >>>
     >>> x = np.arange(12).reshape([6, 2])
     >>> # create an imbalance dataset
@@ -319,7 +319,7 @@ class DataSplitter(SavingMixin):
                 TabularDataset.from_xy(remained_x, remained_y, self._dataset.task_type),
                 self._remained_indices, np.array([], np.int)
             )
-        if n < 1.:
+        if n < 1. or (n == 1. and isinstance(n, float)):
             n = int(round(len(self._x) * n))
         if self.is_time_series:
             split_method = self._split_time_series
@@ -360,6 +360,7 @@ class DataSplitter(SavingMixin):
 class KFold:
     """
     Util class which can perform k-fold data splitting:
+
     1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
     2. Xi ∩ Xj = ∅, ∀ i, j = 1,..., K
     3. X1 ∪ X2 ∪ ... ∪ Xk = X
@@ -381,6 +382,7 @@ class KFold:
     >>> from cfdata.types import np_int_type
     >>> from cfdata.tabular.types import TaskTypes
     >>> from cfdata.tabular.wrapper import TabularDataset
+    >>> from cfdata.tabular.utils import KFold
     >>>
     >>> x = np.arange(12).reshape([6, 2])
     >>> # create an imbalance dataset
@@ -424,6 +426,7 @@ class KFold:
 class KRandom:
     """
     Util class which can perform k-random data splitting:
+
     1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
     2. idx{X1} ≠ idx{X2} ≠ ... ≠ idx{Xk}, where idx{X} = {1, 2, ..., n}
     3. X1 = X2 = ... = Xk = X
@@ -444,6 +447,7 @@ class KRandom:
     >>> from cfdata.types import np_int_type
     >>> from cfdata.tabular.types import TaskTypes
     >>> from cfdata.tabular.wrapper import TabularDataset
+    >>> from cfdata.tabular.utils import KRandom
     >>>
     >>> x = np.arange(12).reshape([6, 2])
     >>> # create an imbalance dataset
@@ -502,6 +506,7 @@ class ImbalancedSampler(LoggingMixin):
     >>>
     >>> from cfdata.types import np_int_type
     >>> from cfdata.tabular import TabularData
+    >>> from cfdata.tabular.utils import ImbalancedSampler
     >>> from cftool.misc import get_counter_from_arr
     >>>
     >>> n = 20
@@ -519,13 +524,15 @@ class ImbalancedSampler(LoggingMixin):
     """
 
     def __init__(self,
-                 data: TabularData,
+                 data,
                  imbalance_threshold: float = 0.1,
                  *,
                  shuffle: bool = True,
                  sample_method: str = "multinomial",
+                 verbose_imbalance: bool = True,
                  verbose_level: int = 2):
         self.data = data
+        self.imbalance_threshold = imbalance_threshold
         self._sample_imbalance_flag = True
         self.shuffle, self._n_samples = shuffle, len(data)
         if not self.shuffle or data.task_type is TaskTypes.REGRESSION:
@@ -547,8 +554,8 @@ class ImbalancedSampler(LoggingMixin):
                 sample_weights /= sample_weights.sum()
                 self._sampler = Sampler(sample_method, sample_weights)
 
-        self._verbose_level = verbose_level
-        if self._sampler is not None:
+        self._sample_method, self._verbose_level = sample_method, verbose_level
+        if verbose_imbalance and self._sampler is not None:
             self.log_msg(
                 f"using imbalanced sampler with label counts = {label_counts.tolist()}",
                 self.info_prefix, 2
@@ -561,6 +568,10 @@ class ImbalancedSampler(LoggingMixin):
     def is_imbalance(self) -> bool:
         return self._sampler is not None
 
+    @property
+    def sample_imbalance(self) -> bool:
+        return self._sample_imbalance_flag
+
     @property
     def label_ratios(self) -> Union[None, np.ndarray]:
         return self._label_ratios
@@ -577,6 +588,16 @@ class ImbalancedSampler(LoggingMixin):
             np.random.shuffle(indices)
         return indices
 
+    def copy(self) -> "ImbalancedSampler":
+        return ImbalancedSampler(
+            self.data,
+            self.imbalance_threshold,
+            shuffle=self.shuffle,
+            sample_method=self._sample_method,
+            verbose_level=self._verbose_level,
+            verbose_imbalance=False
+        )
+
 
 class LabelCollators:
     @staticmethod
@@ -600,6 +621,8 @@ class DataLoader:
     >>>
     >>> from cfdata.types import np_int_type
     >>> from cfdata.tabular import TabularData
+    >>> from cfdata.tabular.utils import DataLoader
+    >>> from cfdata.tabular.utils import ImbalancedSampler
     >>> from cftool.misc import get_counter_from_arr
     >>>
     >>> n = 20
@@ -640,18 +663,17 @@ class DataLoader:
         return n_iter + int(n_iter * self.batch_size < self._n_samples)
 
     def __iter__(self):
-        for attr, init_value in self.reset_caches.items():
-            setattr(self, attr, init_value)
+        self._reset()
         return self
 
     def __next__(self):
         data_next = self._get_next_batch()
         if self._n_siamese == 1:
             return data_next
-        all_data = [data_next] if self._check_valid_batch(data_next) else []
+        all_data = [data_next] if self._check_full_batch(data_next) else []
         while len(all_data) < self._n_siamese:
             data_next = self._get_next_batch()
-            if self._check_valid_batch(data_next):
+            if self._check_full_batch(data_next):
                 all_data.append(data_next)
         x_batch, y_batch = zip(*all_data)
         if self._label_collator is not None:
@@ -659,11 +681,20 @@ class DataLoader:
         return x_batch, y_batch
 
     @property
-    def reset_caches(self):
-        return {
+    def enabled_sampling(self) -> bool:
+        return self.sampler.sample_imbalance
+
+    @enabled_sampling.setter
+    def enabled_sampling(self, value: bool):
+        self.sampler.switch_imbalance_status(value)
+
+    def _reset(self):
+        reset_caches = {
             "_indices_in_use": self.sampler.get_indices(),
             "_siamese_cursor": 0, "_cursor": -1
         }
+        for attr, init_value in reset_caches.items():
+            setattr(self, attr, init_value)
 
     def _get_next_batch(self):
         n_iter, self._cursor = len(self), self._cursor + 1
@@ -679,11 +710,20 @@ class DataLoader:
         batch = self.data[self._indices_in_use[start:end]]
         return batch
 
-    def _check_valid_batch(self, batch):
+    def _check_full_batch(self, batch):
         if len(batch[0]) == self.batch_size:
             return True
         return False
 
+    def copy(self) -> "DataLoader":
+        return DataLoader(
+            self.batch_size,
+            self.sampler.copy(),
+            n_siamese=self._n_siamese,
+            label_collator=self._label_collator,
+            verbose_level=self._verbose_level
+        )
+
 
 __all__ = [
     "SplitResult", "DataSplitter", "KFold", "KRandom",
