@@ -1,15 +1,17 @@
 import os
+import copy
 import logging
 
 import numpy as np
 
 from typing import *
-from cftool.misc import timing_context
+from cftool.misc import timing_context, SavingMixin
 
 from .types import *
 from .recognizer import *
 from .converters import *
 from .processors import *
+from .utils import DataSplitter
 from ..base import DataBase
 from ..types import np_int_type
 
@@ -28,6 +30,8 @@ class TabularData(DataBase):
                  numerical_columns: List[int] = None,
                  categorical_columns: List[int] = None,
                  process_methods: Dict[int, str] = None,
+                 default_numerical_process: str = "normalize",
+                 default_categorical_process: str = "one_hot",
                  label_process_method: str = None,
                  numerical_threshold: float = None,
                  trigger_logging: bool = False,
@@ -52,13 +56,16 @@ class TabularData(DataBase):
         self._numerical_columns = numerical_columns
         self._categorical_columns = categorical_columns
         self._process_methods = process_methods
+        self._default_numerical_process = default_numerical_process
+        self._default_categorical_process = default_categorical_process
         self._label_process_method = label_process_method
         self._numerical_threshold = numerical_threshold
         self._is_file = self._is_arr = False
+        self._num_classes = None
         self._label_idx = self._skip_first = self._delim = None
         self._raw = self._converted = self._processed = None
         self._recognizers = self._converters = self._processors = None
-        self._init_logging(verbose_level, trigger_logging)
+        self._init_logging(verbose_level, trigger=trigger_logging)
         self.excludes = set()
 
     def __len__(self):
@@ -66,7 +73,8 @@ class TabularData(DataBase):
 
     def __getitem__(self, indices: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
         x, y = self.processed.xy
-        return x[indices], y[indices]
+        y_batch = None if y is None else y[indices]
+        return x[indices], y_batch
 
     def __eq__(self, other: "TabularData"):
         if self.raw != other.raw:
@@ -171,31 +179,19 @@ class TabularData(DataBase):
     def is_reg(self) -> bool:
         return self.task_type is TaskTypes.REGRESSION
 
-    # Core
+    @property
+    def is_file(self) -> bool:
+        return self._is_file
 
-    def _read_file(self,
-                   file_path: str) -> Tuple[raw_data_type, raw_data_type]:
-        ext = os.path.splitext(file_path)[1][1:]
-        set_default = lambda n, default: n if n is not None else default
-        if ext == "txt":
-            skip_first, delim = map(set_default, [self._skip_first, self._delim], [False, " "])
-        elif ext == "csv":
-            skip_first, delim = map(set_default, [self._skip_first, self._delim], [True, ","])
-        else:
-            raise NotImplementedError(f"file type '{ext}' not recognized")
-        with open(file_path, "r") as f:
-            if skip_first:
-                f.readline()
-            data = [["nan" if not elem else elem for elem in line.strip().split(delim)] for line in f]
-        if self._label_idx is None:
-            if len(data[0]) != len(self._raw.x[0]):
-                raise ValueError("file contains labels but 'contains_labels=False' passed in")
-            return data, None
-        if self._label_idx < 0:
-            self._label_idx = len(data[0]) + self._label_idx
-        x = [line[:self._label_idx] + line[self._label_idx+1:] for line in data]
-        y = [line[self._label_idx:self._label_idx+1] for line in data]
-        return x, y
+    @property
+    def num_classes(self) -> int:
+        if self.is_reg:
+            return 0
+        if self._num_classes is None:
+            self._num_classes = self._processed.y.max().item() + 1
+        return self._num_classes
+
+    # Core
 
     @staticmethod
     def _flatten(data: data_type) -> data_type:
@@ -272,9 +268,14 @@ class TabularData(DataBase):
                 column_type = self._converters[idx].info.column_type
                 method = None
                 if self._process_methods is not None:
-                    method = self._process_methods.get(idx)
+                    method = self._process_methods.get(idx, "auto")
                 if method is None:
-                    method = "normalize" if column_type is ColumnTypes.NUMERICAL else "one_hot"
+                    method = "identical"
+                elif method == "auto":
+                    if column_type is ColumnTypes.NUMERICAL:
+                        method = self._default_numerical_process
+                    else:
+                        method = self._default_categorical_process
                 processor = self._processors[idx] = processor_dict[method](previous_processors.copy())
                 previous_processors.append(processor)
                 columns = converted_x[..., processor.input_indices]
@@ -302,6 +303,8 @@ class TabularData(DataBase):
             processed_labels = processed_labels.astype(np_int_type)
         self._converted = DataTuple(converted_x, converted_labels)
         self._processed = DataTuple(np.hstack(processed_features), processed_labels)
+        self._valid_columns = [col for col in range(self.raw_dim) if col not in self.excludes]
+        self._valid_columns_dict = None
         return self
 
     def _read_from_file(self,
@@ -313,7 +316,7 @@ class TabularData(DataBase):
         self._is_file = True
         self._label_idx, self._skip_first, self._delim = label_idx, skip_first, delim
         with timing_context(self, "_read_file"):
-            x, y = self._read_file(file_path)
+            x, y = self.read_file(file_path)
         self._raw = DataTuple.with_transpose(x, y)
         return self._core_fit()
 
@@ -324,6 +327,19 @@ class TabularData(DataBase):
         self._raw = DataTuple.with_transpose(x, y)
         return self._core_fit()
 
+    def _transform_labels(self,
+                          raw: DataTuple) -> Tuple[np.ndarray, np.ndarray]:
+        if raw.y is None:
+            converted_labels = transformed_labels = None
+        else:
+            converted_labels = self._converters[-1].convert(self._flatten(raw.y))
+            transformed_labels = self._processors[-1].process(converted_labels.reshape([-1, 1]))
+        if self.task_type is TaskTypes.CLASSIFICATION:
+            if converted_labels is not None:
+                converted_labels = converted_labels.astype(np_int_type)
+                transformed_labels = transformed_labels.astype(np_int_type)
+        return converted_labels, transformed_labels
+
     def _transform(self,
                    raw: DataTuple,
                    return_converted: bool) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
@@ -346,23 +362,50 @@ class TabularData(DataBase):
             idx += processor.input_dim
         transformed_features = np.hstack(processed)
         # transform labels
-        if raw.y is None:
-            converted_labels = transformed_labels = None
-        else:
-            converted_labels = self._converters[-1].convert(self._flatten(raw.y))
-            transformed_labels = self._processors[-1].process(converted_labels.reshape([-1, 1]))
-        # check categorical
-        if self.task_type is TaskTypes.CLASSIFICATION:
-            converted_labels = converted_labels.astype(np_int_type)
-            transformed_labels = transformed_labels.astype(np_int_type)
+        converted_labels, transformed_labels = self._transform_labels(raw)
+        # aggregate
         transformed = DataTuple(transformed_features, transformed_labels)
         if not return_converted:
             return transformed
-        converted = DataTuple(converted_features, converted_labels)
+        converted = DataTuple(converted_features.T, converted_labels)
         return converted, transformed
 
+    def _get_raw(self,
+                 x: Union[str, data_type],
+                 y: data_type = None) -> DataTuple:
+        if self._is_file:
+            if isinstance(x, str):
+                if y is not None:
+                    raise ValueError(f"x refers to file_path but y is still provided ({y}), which is illegal")
+                x, y = self.read_file(x)
+        return DataTuple.with_transpose(x, y)
+
     # API
 
+    def read_file(self,
+                  file_path: str) -> Tuple[raw_data_type, raw_data_type]:
+        ext = os.path.splitext(file_path)[1][1:]
+        set_default = lambda n, default: n if n is not None else default
+        if ext == "txt":
+            skip_first, delim = map(set_default, [self._skip_first, self._delim], [False, " "])
+        elif ext == "csv":
+            skip_first, delim = map(set_default, [self._skip_first, self._delim], [True, ","])
+        else:
+            raise NotImplementedError(f"file type '{ext}' not recognized")
+        with open(file_path, "r") as f:
+            if skip_first:
+                f.readline()
+            data = [["nan" if not elem else elem for elem in line.strip().split(delim)] for line in f]
+        if self._label_idx is None:
+            if len(data[0]) != len(self._raw.x[0]):
+                raise ValueError("file contains labels but 'contains_labels=False' passed in")
+            return data, None
+        if self._label_idx < 0:
+            self._label_idx = len(data[0]) + self._label_idx
+        x = [line[:self._label_idx] + line[self._label_idx+1:] for line in data]
+        y = [line[self._label_idx:self._label_idx+1] for line in data]
+        return x, y
+
     def read(self,
              x: Union[str, data_type],
              y: Union[int, data_type] = -1,
@@ -376,16 +419,50 @@ class TabularData(DataBase):
         self.log_timing()
         return self
 
+    def split(self,
+              n: Union[int, float]) -> Tuple["TabularData", "TabularData"]:
+        splitter = DataSplitter(shuffle=False).fit(self.to_dataset())
+        split = splitter.split(n)
+        split_indices = split.corresponding_indices
+        remained_indices = split.remaining_indices
+        raw, converted, processed = self._raw, self._converted, self._processed
+        p1, p2 = map(copy.copy, [self] * 2)
+        p1._raw, p1._converted, p1._processed = map(
+            DataTuple.split_with,
+            [raw, converted, processed], [split_indices] * 3
+        )
+        p2._raw, p2._converted, p2._processed = map(
+            DataTuple.split_with,
+            [raw, converted, processed], [remained_indices] * 3
+        )
+        return p1, p2
+
+    def copy_to(self,
+                x: Union[str, data_type],
+                y: data_type = None) -> "TabularData":
+        copied = copy.copy(self)
+        raw = copied._raw = self._get_raw(x, y)
+        copied._converted, copied._processed = self._transform(raw, True)
+        return copied
+
     def transform(self,
                   x: Union[str, data_type],
                   y: data_type = None,
                   *,
                   return_converted: bool = False) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
-        if self._is_file:
-            x, y = self._read_file(x)
-        raw = DataTuple.with_transpose(x, y)
+        raw = self._get_raw(x, y)
         return self._transform(raw, return_converted)
 
+    def transform_labels(self,
+                         y: data_type,
+                         *,
+                         return_converted: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
+        raw = DataTuple(None, y)
+        converted_labels, transformed_labels = self._transform_labels(raw)
+        if not return_converted:
+            return transformed_labels
+        return converted_labels, transformed_labels
+
     def recover_labels(self,
                        y: np.ndarray,
                        *,
@@ -394,12 +471,18 @@ class TabularData(DataBase):
         convert_recovered = self.converters[-1].recover(process_recovered.ravel(), inplace=inplace)
         return convert_recovered.reshape([-1, 1])
 
-    def load(self, folder, *, compress=True) -> "TabularData":
-        super().load(folder)
-        is_file, is_arr = self._is_file, self._is_arr
-        self.read(*self._raw[:2])
-        self._is_file, self._is_arr = is_file, is_arr
-        return self
+    @classmethod
+    def load(cls,
+             folder: str,
+             *,
+             compress: bool = True,
+             verbose_level: int = 0) -> "TabularData":
+        data = cls(verbose_level=verbose_level)
+        SavingMixin.load(data, folder, compress=compress)
+        is_file, is_arr = data._is_file, data._is_arr
+        data.read(*data._raw[:2])
+        data._is_file, data._is_arr = is_file, is_arr
+        return data
 
     def to_dataset(self) -> TabularDataset:
         return TabularDataset(*self.processed.xy, task_type=self.task_type)
